"""
"""

import os
import copy
import logging
import re
import json
import jsonschema

from typing import Dict, Optional, List, Tuple, Any, Callable, Union
from pydantic import BaseModel

# from typing import Dict, Optional, List, Tuple, Any
from collections import defaultdict, OrderedDict
from lxml import etree
import xmljson

from label_studio_sdk._legacy.exceptions import (
    LSConfigParseException,
    LabelStudioXMLSyntaxErrorSentryIgnored,
    LabelStudioValidationErrorSentryIgnored,
)

from .control_tags import (
    ControlTag,
    ChoicesTag,
    LabelsTag,
)
from .object_tags import ObjectTag
from .label_tags import LabelTag
from label_studio_sdk._legacy.objects import AnnotationValue, TaskValue, PredictionValue


dir_path = os.path.dirname(os.path.realpath(__file__))
file_path = os.path.join(dir_path, "..", "_legacy", "schema", "label_config_schema.json")

with open(file_path) as f:
    _LABEL_CONFIG_SCHEMA_DATA = json.load(f)

_LABEL_TAGS = {"Label", "Choice", "Relation"}

_DIR_APP_NAME = "label-studio"
_VIDEO_TRACKING_TAGS = {"videorectangle"}

RESULT_KEY = "result"

############ core/label_config.py


def merge_labels_counters(dict1, dict2):
    """
    Merge two dictionaries with nested dictionary values into a single dictionary.

    Args:
        dict1 (dict): The first dictionary to merge.
        dict2 (dict): The second dictionary to merge.

    Returns:
        dict: A new dictionary with the merged nested dictionaries.

    Example:
        dict1 = {'sentiment': {'Negative': 1, 'Positive': 1}}
        dict2 = {'sentiment': {'Positive': 2, 'Neutral': 1}}
        result_dict = merge_nested_dicts(dict1, dict2)
        # {'sentiment': {'Negative': 1, 'Positive': 3, 'Neutral': 1}}
    """
    result_dict = {}

    # iterate over keys in both dictionaries
    for key in set(dict1.keys()) | set(dict2.keys()):
        # add the corresponding values if they exist in both dictionaries
        value = {}
        if key in dict1:
            value.update(dict1[key])
        if key in dict2:
            for subkey in dict2[key]:
                value[subkey] = value.get(subkey, 0) + dict2[key][subkey]
        # add the key-value pair to the result dictionary
        result_dict[key] = value

    return result_dict


def _fix_choices(config):
    """
    workaround for single choice
    https://github.com/heartexlabs/label-studio/issues/1259
    """
    if "Choices" in config:
        # for single Choices tag in View
        if "Choice" in config["Choices"] and not isinstance(
            config["Choices"]["Choice"], list
        ):
            config["Choices"]["Choice"] = [config["Choices"]["Choice"]]
        # for several Choices tags in View
        elif isinstance(config["Choices"], list) and all(
            "Choice" in tag_choices for tag_choices in config["Choices"]
        ):
            for n in range(len(config["Choices"])):
                # check that Choices tag has only 1 choice
                if not isinstance(config["Choices"][n]["Choice"], list):
                    config["Choices"][n]["Choice"] = [config["Choices"][n]["Choice"]]
    if "View" in config:
        if isinstance(config["View"], OrderedDict):
            config["View"] = _fix_choices(config["View"])
        else:
            config["View"] = [_fix_choices(view) for view in config["View"]]
    return config


def get_annotation_tuple(from_name, to_name, type):
    if isinstance(to_name, list):
        to_name = ",".join(to_name)
    return "|".join([from_name, to_name, type.lower()])


def get_all_control_tag_tuples(label_config):
    # "chc|text|choices"
    outputs = parse_config(label_config)
    out = []
    for control_name, info in outputs.items():
        out.append(get_annotation_tuple(control_name, info["to_name"], info["type"]))
    return out


def get_all_types(label_config):
    """
    Get all types from label_config
    """
    outputs = parse_config(label_config)
    out = []
    for control_name, info in outputs.items():
        out.append(info["type"].lower())
    return out


def display_count(count: int, type: str) -> Optional[str]:
    """Helper for displaying pluralized sources of validation errors,
    eg "1 draft" or "3 annotations"
    """
    if not count:
        return None

    return f'{count} {type}{"s" if count > 1 else ""}'


######################


class LabelInterface:
    """The LabelInterface class serves as an interface to parse and
    validate labeling configurations, annotations, and predictions
    within the Label Studio ecosystem.

    It is designed to be compatible at the data structure level with
    an existing parser widely used within the Label Studio ecosystem.
    This ensures that it works seamlessly with most of the existing functions,
    either by directly supporting them or by offering re-implemented versions
    through the new interface.

    Moreover, the parser adds value by offering functionality to
    validate predictions and annotations against the specified
    labeling configuration. Below is a simple example of how to use
    the new API:

    ```python
    from label_studio_sdk.label_interface import LabelInterface

    config = "<View><Text name='txt' value='$val' /><Choices name='chc' toName='txt'><Choice value='one'/> <Choice value='two'/></Choices></View>"

    li = LabelInterface(config)
    region = li.get_tag("chc").label("one")

    # returns a JSON representing a Label Studio region
    region.as_json()

    # returns True
    li.validate_prediction({
      "model_version": "0.0.1",
      "score": 0.90,
      "result": [{
          "from_name": "chc",
          "to_name": "txt",
          "type": "choices",
          "value": { "choices": ["one"] }
      }]
    })
    ```
    """

    def __init__(self, config: str, tags_mapping=None, *args, **kwargs):
        """
        Initialize a LabelInterface instance using a config string.

        Args:
          config (str): Configuration string.
          tags_mapping: Provide your own implementation of any tag, this is helpful in cases where you want to override one of the control tags, and have your custom .label() method implemented.

        The configuration string should be formatted as follows:

        ```
        <View>
          <Choices name="sentiment" toName="txt">
            <Choice value="Positive" />
            <Choice value="Negative" />
            <Choice value="Neutral" />
          </Choices>
          <Text name="txt" value="$text" />
        </View>
        ```
        This method will extract the predefined task from the configuration and 
        parse the controls, objects, and labels used in it.
        """
        self.task_loaded = False
        
        self._config = config
        self._tags_mapping = tags_mapping
        
        # extract predefined task from the config
        _task_data, _ann, _pred = LabelInterface.get_task_from_labeling_config(config)
        self._sample_config_task = _task_data
        self._sample_config_ann = _ann
        self._sample_config_pred = _pred

        controls, objects, labels, tree = self.parse(config)
        controls = self._link_controls(controls, objects, labels)

        # list of control tags that this config has
        self._control_tags = set(controls.keys())
        self._object_tags = set(objects.keys())
        # self._label_names = set(labels.keys())

        self._controls = controls
        self._objects = objects
        self._labels = labels
        self._tree = tree

        

    ##### NEW API

    @property
    def controls(self):
        """Returns list of control tags"""
        return self._controls and self._controls.values()

    @property
    def objects(self):
        """Returns list of object tags"""
        return self._objects and self._objects.values()

    @property
    def labels(self):
        """Returns list of label tags"""
        return self._labels and self._labels.values()

    def _link_controls(self, controls: Dict, objects: Dict, labels: Dict) -> Dict:
        """ """
        for name, tag in controls.items():
            inputs = []
            for object_tag_name in tag.to_name:
                if object_tag_name not in objects:
                    # logger.info(
                    #     f'to_name={object_tag_name} is specified for output tag name={name}, '
                    #     'but we can\'t find it among input tags'
                    # )
                    continue

                inputs.append(objects[object_tag_name])

            tag.set_objects(inputs)
            tag.set_labels(list(labels[name]))
            tag.set_labels_attrs(labels[name])

        return controls

    def _get_tag(self, name, tag_store):
        """ """
        if name is not None:
            if name not in tag_store:
                raise Exception(
                    f"Name {name} is not found, available names: {tag_store.keys()}"
                )
            else:
                return tag_store[name]

        if tag_store and len(tag_store.keys()) > 1:
            raise Exception("Multiple object tags connected, you should specify name")

        return list(tag_store.values())[0]

    def get_tag(self, name):
        """Method to retrieve the tag object by its name from the current instance.

        The method checks if the tag with the provided name exists in
        either `_controls` or `_objects` attributes of the current
        instance.  If a match is found, it returns the tag. If the tag
        is not found an exception is raised.

        Args:
            name (str): Name of the tag to be retrieved.

        Returns:
            object: The tag object if it exists in either `_controls` or `_objects`.

        Raises:
            Exception: If the tag with the given name does not exist in both `_controls` and `_objects`.

        """
        if name in self._controls:
            return self._controls[name]

        if name in self._objects:
            return self._objects[name]

        raise Exception(f"Tag with name {name} not found")

    def get_object(self, name=None):
        """Retrieves the object with the given name from `_objects`.

        This utilizes the `_get_tag` method to obtain the named object.

        Args:
            name (str, optional): The name of the object to be retrieved from `_objects`.

        Returns: object: The corresponding object if it exists in
            `_objects`.

        """
        return self._get_tag(name, self._objects)

    def get_output(self, name=None):
        """Provides an alias for the `get_control` method."""
        return self.get_control(name)

    def get_control(self, name=None):
        """Retrieves the control tag that the control tag maps to.

        This uses the `_get_tag` method to obtain the named control.

        Args:
            name (str, optional): The name of the control to be retrieved.

        Returns: object: The corresponding control if it exists in
            `_controls`.

        """
        return self._get_tag(name, self._controls)

    def find_tags_by_class(self, tag_class) -> List:
        """Finds tags by their class type.

        The function looks into both `self.objects` and
        `self.controls` to find tags that are instances of the
        provided class(es)

        Args:
            tag_class (class or list of classes): The class type(s) of the tags to be found.

        Returns:
            list: A list of tags that are instances of the provided `tag_class`(es).

        """
        lst = list(self.objects) + list(self.controls)
        tag_classes = [tag_class] if not isinstance(tag_class, list) else tag_class

        return [tag for tag in lst for cls in tag_classes if isinstance(tag, cls)]

    def find_tags(
        self, tag_type: Optional[str] = None, match_fn: Optional[Callable] = None
    ) -> List:
        """Finds tags that match the given function in the entire parsed tree.

        This function searches through both `objects` and `controls`
        based on `tag_type`, and applies the `match_fn` (if provided)
        to filter matching tags.

        Args:
            tag_type (str, optional): The type of tags to be
                searched. Categories include 'objects', 'controls',
                'inputs' (alias for 'objects'), 'outputs' (alias for
                'controls'). If not specified, searches both 'objects'
                and 'controls'.
            match_fn (Callable, optional): A function that takes a tag
                as an input and returns a boolean value indicating
                whether the tag matches the required condition.

        Returns: list: A list of tags that match the given type and
        satisfy `match_fn`.

        """
        tag_types = {
            "objects": self.objects,
            "controls": self.controls,
            # aliases
            "inputs": self.objects,
            "outputs": self.controls,
        }

        lst = tag_types.get(tag_type, list(self.objects) + list(self.controls))

        if match_fn is not None:
            lst = list(filter(match_fn, lst))

        return lst
    
    def load_task(self, task):
        """Loads a task and substitutes the value in each object tag
        with actual data from the task, returning a copy of the
        LabelConfig object.

        If the `value` field in an object tag is designed to take
        variable input (i.e., `value_is_variable` is True), the
        function replaces this value with the corresponding value from
        the task dictionary.

        Args:
            task (dict): Dictionary representing the task, where
            each key-value pair denotes an attribute-value of the
            task.

        Returns:
            LabelInterface: A deep copy of the current LabelIntreface
            instance with the object tags' value fields populated with
            data from the task.

        """
        tree = copy.deepcopy(self)
        tree.task_loaded = True
        
        for obj in tree.objects:
            print(obj.value_is_variable, obj.value_name)
            if obj.value_is_variable and obj.value_name in task:
                obj.value = task.get(obj.value_name)

        return tree
    
    def parse(self, config_string: str) -> Tuple[Dict, Dict, Dict, etree._Element]:
        """Parses the received configuration string into dictionaries
        of ControlTags, ObjectTags, and Labels, along with an XML tree
        of the configuration.

        Args:
            config_string (str): the configuration string to be parsed.

        Returns:
            Tuple of:
            - Dictionary where keys are control tag names and values are ControlTag instances.
            - Dictionary where keys are object tag names and values are ObjectTag instances.
            - Dictionary of dictionaries where primary keys are label parent names
              and secondary keys are label values and values are LabelTag instances.
            - An XML tree of the configuration.
        """
        try:
            xml_tree = etree.fromstring(config_string)
        except etree.XMLSyntaxError as e:
            raise LabelStudioXMLSyntaxErrorSentryIgnored(str(e))

        objects, controls, labels = {}, {}, defaultdict(dict)

        variables = []

        for tag in xml_tree.iter():
            if tag.attrib and "indexFlag" in tag.attrib:
                variables.append(tag.attrib["indexFlag"])

            if ControlTag.validate_node(tag):
                controls[tag.attrib["name"]] = ControlTag.parse_node(tag, tags_mapping=self._tags_mapping)

            elif ObjectTag.validate_node(tag):
                objects[tag.attrib["name"]] = ObjectTag.parse_node(tag, tags_mapping=self._tags_mapping)

            elif LabelTag.validate_node(tag):
                lb = LabelTag.parse_node(tag, controls)
                # This case is hit when Label tag is missing `value` and `alias`
                # For now we will skip that Label, but in future might want to raise an error
                if lb:
                    labels[lb.parent_name][lb.value] = lb

        return controls, objects, labels, xml_tree

    @classmethod
    def parse_config_to_json(cls, config_string):
        """ """
        try:
            xml = etree.fromstring(config_string)
        except TypeError:
            raise etree.ParseError("can only parse strings")
        if xml is None:
            raise etree.ParseError("xml is empty or incorrect")

        config = xmljson.badgerfish.data(xml)
        config = _fix_choices(config)

        return config

    def _schema_validation(self, config_string):
        """ """
        try:
            config = LabelInterface.parse_config_to_json(config_string)
            jsonschema.validate(config, _LABEL_CONFIG_SCHEMA_DATA)
        except (etree.ParseError, ValueError) as exc:
            raise LabelStudioValidationErrorSentryIgnored(str(exc))
        except jsonschema.exceptions.ValidationError as exc:
            error_message = exc.context[-1].message if len(exc.context) else exc.message
            error_message = "Validation failed on {}: {}".format(
                "/".join(map(str, exc.path)), error_message.replace("@", "")
            )
            raise LabelStudioValidationErrorSentryIgnored(error_message)

    def _to_name_validation(self, config_string):
        """ """
        # toName points to existent name
        all_names = re.findall(r'name="([^"]*)"', config_string)

        names = set(all_names)
        toNames = re.findall(r'toName="([^"]*)"', config_string)
        for toName_ in toNames:
            for toName in toName_.split(","):
                if toName not in names:
                    raise LabelStudioValidationErrorSentryIgnored(
                        f'toName="{toName}" not found in names: {sorted(names)}'
                    )

    def _unique_names_validation(self, config_string):
        """ """
        # unique names in config # FIXME: 'name =' (with spaces) won't work
        all_names = re.findall(r'name="([^"]*)"', config_string)
        if len(set(all_names)) != len(all_names):
            raise LabelStudioValidationErrorSentryIgnored(
                "Label config contains non-unique names"
            )

    @property
    def is_valid(self):
        """ """
        try:
            self.validate()
            return True
        except LabelStudioValidationErrorSentryIgnored:
            return False

    def validate(self):
        """Validates the provided configuration string against various validation criteria.

        This method applies a series of validation checks to
        `_config`, including schema validation, checking for
        uniqueness of names used in the configuration, and the
        "to_name" validation. It throws exceptions if any of these
        validations fail.

        Raises:
            Exception: If any validation fails, specific to the type of validation.

        """
        config_string = self._config

        self._schema_validation(config_string)
        self._unique_names_validation(config_string)
        self._to_name_validation(config_string)

    @classmethod
    def validate_with_data(cls, config):
        """ """
        raise NotImplemented()

    def validate_task(self, task: "TaskValue", validate_regions_only=False):
        """ """
        # TODO this might not be always true, and we need to use
        # "strict" param above to be able to configure

        # for every object tag we've got that has value as it's
        # variable we need to have an associated item in the task data
        for obj in self.objects:
            if obj.value_is_variable and task["data"].get(obj.value_name, None) is None:
                return False

        if "annotations" in task and not self.validate_annotation():
            return False

        if "predictions" in task and not self.validate_prediction():
            return False

        return True

    def _validate_object(self, obj):
        """ """
        regions = []
        for r in obj.get(RESULT_KEY):
            if r.get('type') != "relation":
                if not self.validate_region(r):
                    return False

                regions.append(r)

        for r in obj.get(RESULT_KEY):
            if r.get('type') == "relation" and \
               not self.validate_relation(r, regions):
                return False

        return True                                                
        
    def validate_annotation(self, annotation):
        """Validates the given annotation against the current configuration.

        This method applies the `validate_region` method to each
        region in the annotation and returns False if any of these
        validations fail. If all the regions pass the validation, it
        returns True.

        Args:
            annotation (dict): The annotation to be validated, where
            each key-value pair denotes an attribute-value of the
            annotation.

        Returns:
            bool: True if all regions in the annotation pass the
            validation, False otherwise.

        """
        return self._validate_object(annotation)

    def validate_prediction(self, prediction):
        """Same as validate_annotation right now"""
        return self._validate_object(prediction)

    def validate_region(self, region) -> bool:
        """Validates a region from the annotation against the current
        configuration.

        The validation checks the following:
        - Both control and object items are present in the labeling configuration.
        - The type of the region matches the control tag name.
        - The 'to_name' in the region data connects to the same tag as in the configuration.
        - The actual value for example in <Labels /> tag is producing start, end, and labels.

        If any of these validations fail, the function immediately
        returns False. If all validations pass for a region, it
        returns True.

        Args:
            region (dict): The region to be validated.

        Returns:
            bool: True if all checks pass for the region, False otherwise.

        """
        control = self.get_control(region["from_name"])
        obj = self.get_object(region["to_name"])

        # we should have both items present in the labeling config
        if not control or not obj:
            return False

        # type of the region should match the tag name        
        if control.tag.lower() != region["type"]:
            return False
        
        # make sure that in config it connects to the same tag as
        # immplied by the region data
        if region["to_name"] not in control.to_name:
            return False
        
        # validate the actual value, for example that <Labels /> tag
        # is producing start, end, and labels
        if not control.validate_value(region["value"]):
            return False
        
        return True

    def validate_relation(self, relation, regions, _mapping=None) -> bool:
        """Validates that the relation is correct and all the associated objects are provided"""
        if _mapping is None:
            _mapping = { r['id']: r for r in regions }

        if relation.get("type") != "relation" or \
           relation.get("direction") not in ("left", "right", "bi") or \
           relation.get("from_id") not in _mapping or \
           relation.get("to_id") not in _mapping:
            return False
        
        return True
    
    ### Generation

    def _sample_task(self, secure_mode=False):
        """ """
        # predefined_task, annotations, predictions = get_task_from_labeling_config(label_config)
        generated_task = self.generate_sample_task(
            mode="editor_preview", secure_mode=secure_mode
        )

        if self._sample_config_task is not None:
            generated_task.update(self._sample_config_task)

        return generated_task, self._sample_config_ann, self._sample_config_pred

    def generate_sample_task(self, mode="upload", secure_mode=False):
        """Generates a sample task based on the provided mode and
        secure_mode.

        This function generates an example value for each object in
        `self.objects` using the specified `mode` and
        `secure_mode`. The resulting task is a dictionary where each
        key-value pair denotes an object's value-name and example
        value.

        Args:
            mode (str, optional): The operation mode. Accepts any string but defaults to 'upload'.
            secure_mode (bool, optional): The security mode. Defaults to False.

        Returns:
            dict: A dictionary representing the sample task.

        """
        task = {
            obj.value_name: obj.generate_example_value(
                mode=mode, secure_mode=secure_mode
            )
            for obj in self.objects
        }

        return task

    def generate_sample_annotation(self):
        """ """
        raise NotImplemented()

    #####
    ##### COMPATIBILITY LAYER
    #####
    ##### This are re-implmenetation of functions found in different
    ##### label_config.py files across the repo. Not all of this were
    ##### tested, therefore I suggest to write a test first, and then
    ##### replace where it's being used in the repo.

    def config_essential_data_has_changed(self, new_config_str):
        """Detect essential changes of the labeling config"""
        new_obj = LabelInterface(config=new_config_str)

        for new_tag_name, new_tag in new_obj._controls.items():
            if new_tag_name not in self._controls:
                return True

            old_tag = self._controls[new_tag_name]

            if new_tag.tag != old_tag.tag:
                return True
            if new_tag.objects != old_tag.objects:
                return True
            if not set(old_tag.labels).issubset(new_tag.labels):
                return True

        return False

    def generate_sample_task_without_check(
        label_config, mode="upload", secure_mode=False
    ):
        """ """
        raise NotImplemented()

    @classmethod
    def get_task_from_labeling_config(cls, config):
        """Get task, annotations and predictions from labeling config comment,
        it must start from "<!-- {" and end as "} -->"
        """
        # try to get task data, annotations & predictions from config comment
        task_data, annotations, predictions = {}, None, None
        start = config.find("<!-- {")
        start = start if start >= 0 else config.find("<!--{")
        start += 4
        end = config[start:].find("-->") if start >= 0 else -1

        if 3 < start < start + end:
            try:
                # logger.debug('Parse ' + config[start : start + end])
                body = json.loads(config[start : start + end])
            except Exception:
                # logger.error("Can't parse task from labeling config", exc_info=True)
                pass
            else:
                # logger.debug(json.dumps(body, indent=2))
                dont_use_root = "predictions" in body or "annotations" in body
                task_data = (
                    body["data"]
                    if "data" in body
                    else (None if dont_use_root else body)
                )
                predictions = body["predictions"] if "predictions" in body else None
                annotations = body["annotations"] if "annotations" in body else None

        return task_data, annotations, predictions

    @classmethod
    def config_line_stipped(self, c):
        tree = etree.fromstring(c, forbid_dtd=False)
        comments = tree.xpath("//comment()")

        for c in comments:
            p = c.getparent()
            if p is not None:
                p.remove(c)
            c = etree.tostring(tree, method="html").decode("utf-8")

        return c.replace("\n", "").replace("\r", "")

    def get_all_control_tag_tuples(label_config):
        """ """
        return [tag.as_tuple() for tag in self.controls]

    def get_first_tag_occurence(
        self,
        control_type: Union[str, Tuple],
        object_type: Union[str, Tuple],
        name_filter: Optional[Callable] = None,
        to_name_filter: Optional[Callable] = None,
    ) -> Tuple[str, str, str]:
        """
        Reads config and fetches the first control tag along with first object tag that matches the type.

        Args:
          control_type (str or tuple): The control type for checking tag matches.
          object_type (str or tuple): The object type for checking tag matches.
          name_filter (function, optional): If given, only tags with this name will be considered.
                                           Default is None.
          to_name_filter (function, optional): If given, only tags with this name will be considered.
                                              Default is None.

        Returns:
          tuple: (from_name, to_name, value), representing control tag, object tag and input value.
        """

        for tag in self.controls:
            if tag.match(control_type, name_filter_fn=name_filter):
                for object_tag in tag.objects:
                    if object_tag.match(object_type, to_name_filter_fn=to_name_filter):
                        return tag.name, object_tag.name, object_tag.value_name

        raise ValueError(
            f"No control tag of type {control_type} and object tag of type {object_type} found in label config"
        )

    def get_all_labels(self):
        """ """
        dynamic_values = {c.name: True for c in self.controls if c.dynamic_value}
        return self._labels, dynamic_values

    def get_all_object_tag_names(self):
        """ """
        return self._objects.keys()

    def extract_data_types(self):
        """ """
        return self._objects

    def is_video_object_tracking(self):
        """ """
        match_fn = lambda tag: tag.tag.lower() in _VIDEO_TRACKING_TAGS
        tags = self.find_tags(match_fn=match_fn)

        return bool(tags)

    def is_type(self, tag_type=None):
        """ """
        raise NotImplemented

    # NOTE: you can use validate() instead
    # def validate_label_config(self, config_string):
    #     # xml and schema
    #     self._schema_validation(config_string)
    #     self._unique_names_validation(config_string)
    #     self._to_name_validation(config_string)

    def validate_config_using_summary(self, summary, strict=False):
        """Validate current config using LS Project Summary"""
        # this is a rewrite of project.validate_config function
        # self.validate_label_config(config_string)
        if not self._objects:
            return False

        created_labels = summary.created_labels
        created_labels_drafts = summary.created_labels_drafts
        annotations_summary = summary.created_annotations

        self.validate_annotations_consistency(annotations_summary)
        self.validate_lables_consistency(created_labels, created_labels_drafts)

    def validate_lables_consistency(self, created_labels, created_labels_drafts):
        """ """
        # validate labels consistency
        # labels_from_config, dynamic_values_tags = self.get_all_labels(config_string)

        created_labels = merge_labels_counters(created_labels, created_labels_drafts)

        # <Labels name="sentinement" ...><Label value="Negative" ... />
        # {'sentiment': {'Negative': 1, 'Positive': 3, 'Neutral': 1}}

        for control_tag_from_data, labels_from_data in created_labels.items():
            # Check if labels created in annotations, and their control tag has been removed
            control_from_config = self.get_control(control_tag_from_data)

            if labels_from_data and not control_from_config:
                raise LabelStudioValidationErrorSentryIgnored(
                    f"There are {sum(labels_from_data.values(), 0)} annotation(s) created with tag "
                    f'"{control_tag_from_data}", you can\'t remove it'
                )

            removed_labels = []
            # Check that labels themselves were not removed
            for label_name, label_value in labels_from_data.items():
                if label_value > 0 and not control_from_config.labels_attrs.get(
                    label_name, None
                ):
                    # that label was used in labeling before, but not
                    # present in the current config
                    removed_labels.append(label_name)

            # TODO that needs to be added back
            # if 'VideoRectangle' in tag_types:
            #     for key in labels_from_config:
            #         labels_from_config_by_tag |= set(labels_from_config[key])

            # if 'Taxonomy' in tag_types:
            #     custom_tags = Label.objects.filter(links__project=self).values_list('value', flat=True)
            #     flat_custom_tags = set([item for sublist in custom_tags for item in sublist])
            #     labels_from_config_by_tag |= flat_custom_tags

            if len(removed_labels):
                raise LabelStudioValidationErrorSentryIgnored(
                    f'These labels still exist in annotations or drafts:\n{",".join(removed_labels)}'
                    f'Please add labels to tag with name="{str(control_tag_from_data)}".'
                )

    def validate_annotations_consistency(self, annotations_summary):
        """ """
        # annotations_summary is coming from LS Project Summary, it's
        # format is: { "chc|text|choices": 10 }
        # which means that there are two tags, Choices, and one of
        # object tags and there are 10 annotations

        err = []
        annotations_from_data = set(annotations_summary)

        for ann in annotations_from_data:
            from_name, to_name, tag_type = ann.split("|")

            # avoid textarea to_name check (see DEV-1598)
            if tag_type.lower() == "textarea":
                continue

            try:
                control = self.get_control(from_name)
                if not control or not control.get_object(to_name):
                    err.append(
                        f"with from_name={from_name}, to_name={to_name}, type={tag_type}"
                    )
            except Exception as ex:
                err.append(
                    f"Error occurred while processing from_name={from_name}, to_name={to_name}, type={tag_type}, error: {str(ex)}"
                )

            # control = self.get_control(from_name)
            # if not control or not control.get_object(to_name):
            #     err.append(f'with from_name={from_name}, to_name={to_name}, type={tag_type}')

        if err:
            diff_str = "\n".join(err)
            raise LabelStudioValidationErrorSentryIgnored(
                f"Created annotations are incompatible with provided labeling schema, we found:\n{diff_str}"
            )



"""
"""

import json
from uuid import uuid4

from typing import Any, List, Dict, Optional
from pydantic import BaseModel, Field


class Region(BaseModel):
    """
    Class for Region Tag

    Attributes:
    -----------
    id: str
        The unique identifier of the region
    x: int
        The x coordinate of the region
    y: int

    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    from_tag: Any
    to_tag: Any
    value: Any
    relations: Optional[List[Dict]] = []

    def _dict(self):
        """ """
        return {
            "id": self.id,
            "from_name": self.from_tag.name,
            "to_name": self.to_tag.name,
            "type": self.from_tag.tag.lower(),
            # TODO This needs to be improved
            "value": self.value.dict(),
        }

    def _dict_relations(self):
        """ """
        # this code does not include "labels" key if no labels were passed
        return [
            {**{
                "from_id": self.id,
                "to_id": rel.get("region", {}).id,
                "type": "relation",
                "direction": rel.get("direction", "right")},
             **({"labels": rel["labels"]} if rel.get("labels") else {})
             }
            for rel in self.relations
        ]
               
    def to_json(self):
        """ """
        return json.dumps(self._dict())

    def to_json_relations(self):
        """ """
        return json.dumps(self._dict_relations())

    @property
    def has_relations(self):
        return len(self.relations) > 0
    
    def add_relation(self, region=None, direction="right", label=None):
        """ """
        self.relations.append({ "region": region, "direction": direction, "labels": label })
        
    def set_relations(self, rels):
        """ """
        self.relations = rels
        



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import pydantic_v1
from ..core.request_options import RequestOptions
from ..types.base_user import BaseUser
from .types.users_get_token_response import UsersGetTokenResponse
from .types.users_reset_token_response import UsersResetTokenResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class UsersClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def reset_token(self, *, request_options: typing.Optional[RequestOptions] = None) -> UsersResetTokenResponse:
        """
        Reset your access token or API key. When reset, any scripts or automations you have in place will need to be updated with the new key.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UsersResetTokenResponse
            User token response

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.reset_token()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/current-user/reset-token/", method="POST", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UsersResetTokenResponse, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def get_token(self, *, request_options: typing.Optional[RequestOptions] = None) -> UsersGetTokenResponse:
        """
        Get a access token to authenticate to the API as the current user. To find this in the Label Studio interface, click **Account & Settings** in the upper right. For more information, see [Access Token](https://labelstud.io/guide/user_account#Access-token).

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UsersGetTokenResponse
            User token response

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.get_token()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/current-user/token", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UsersGetTokenResponse, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def whoami(self, *, request_options: typing.Optional[RequestOptions] = None) -> BaseUser:
        """
        Get information about your user account, such as your username, email, and user ID.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.whoami()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/current-user/whoami", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def list(self, *, request_options: typing.Optional[RequestOptions] = None) -> typing.List[BaseUser]:
        """
        List all users in your Label Studio organization.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[BaseUser]


        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.list()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/users/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(typing.List[BaseUser], _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def create(
        self,
        *,
        id: typing.Optional[int] = OMIT,
        first_name: typing.Optional[str] = OMIT,
        last_name: typing.Optional[str] = OMIT,
        username: typing.Optional[str] = OMIT,
        email: typing.Optional[str] = OMIT,
        avatar: typing.Optional[str] = OMIT,
        initials: typing.Optional[str] = OMIT,
        phone: typing.Optional[str] = OMIT,
        allow_newsletters: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseUser:
        """
        Create a user in Label Studio.

        Parameters
        ----------
        id : typing.Optional[int]
            User ID

        first_name : typing.Optional[str]
            First name of the user

        last_name : typing.Optional[str]
            Last name of the user

        username : typing.Optional[str]
            Username of the user

        email : typing.Optional[str]
            Email of the user

        avatar : typing.Optional[str]
            Avatar URL of the user

        initials : typing.Optional[str]
            Initials of the user

        phone : typing.Optional[str]
            Phone number of the user

        allow_newsletters : typing.Optional[bool]
            Whether the user allows newsletters

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.create()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/users/",
            method="POST",
            json={
                "id": id,
                "first_name": first_name,
                "last_name": last_name,
                "username": username,
                "email": email,
                "avatar": avatar,
                "initials": initials,
                "phone": phone,
                "allow_newsletters": allow_newsletters,
            },
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def get(self, id: int, *, request_options: typing.Optional[RequestOptions] = None) -> BaseUser:
        """
        Get info about a specific Label Studio user.
        You will need to provide their user ID. You can find a list of all user IDs using [List users](list).

        Parameters
        ----------
        id : int
            User ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.get(
            id=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/users/{jsonable_encoder(id)}/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def delete(self, id: int, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete a specific Label Studio user.

        You will need to provide their user ID. You can find a list of all user IDs using [List users](list).

        <Warning>Use caution when deleting a user, as this can cause issues such as breaking the "Annotated by" filter or leaving orphaned records.</Warning>

        Parameters
        ----------
        id : int
            User ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.delete(
            id=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/users/{jsonable_encoder(id)}/", method="DELETE", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def update(
        self,
        id: int,
        *,
        users_update_request_id: typing.Optional[int] = OMIT,
        first_name: typing.Optional[str] = OMIT,
        last_name: typing.Optional[str] = OMIT,
        username: typing.Optional[str] = OMIT,
        email: typing.Optional[str] = OMIT,
        avatar: typing.Optional[str] = OMIT,
        initials: typing.Optional[str] = OMIT,
        phone: typing.Optional[str] = OMIT,
        allow_newsletters: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseUser:
        """
        Update details for a specific Label Studio user, such as their name or contact information.

        You will need to provide their user ID. You can find a list of all user IDs using [List users](list).

        Parameters
        ----------
        id : int
            User ID

        users_update_request_id : typing.Optional[int]
            User ID

        first_name : typing.Optional[str]
            First name of the user

        last_name : typing.Optional[str]
            Last name of the user

        username : typing.Optional[str]
            Username of the user

        email : typing.Optional[str]
            Email of the user

        avatar : typing.Optional[str]
            Avatar URL of the user

        initials : typing.Optional[str]
            Initials of the user

        phone : typing.Optional[str]
            Phone number of the user

        allow_newsletters : typing.Optional[bool]
            Whether the user allows newsletters

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.users.update(
            id=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/users/{jsonable_encoder(id)}/",
            method="PATCH",
            json={
                "id": users_update_request_id,
                "first_name": first_name,
                "last_name": last_name,
                "username": username,
                "email": email,
                "avatar": avatar,
                "initials": initials,
                "phone": phone,
                "allow_newsletters": allow_newsletters,
            },
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncUsersClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def reset_token(self, *, request_options: typing.Optional[RequestOptions] = None) -> UsersResetTokenResponse:
        """
        Reset your access token or API key. When reset, any scripts or automations you have in place will need to be updated with the new key.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UsersResetTokenResponse
            User token response

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.reset_token()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/current-user/reset-token/", method="POST", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UsersResetTokenResponse, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def get_token(self, *, request_options: typing.Optional[RequestOptions] = None) -> UsersGetTokenResponse:
        """
        Get a access token to authenticate to the API as the current user. To find this in the Label Studio interface, click **Account & Settings** in the upper right. For more information, see [Access Token](https://labelstud.io/guide/user_account#Access-token).

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UsersGetTokenResponse
            User token response

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.get_token()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/current-user/token", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UsersGetTokenResponse, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def whoami(self, *, request_options: typing.Optional[RequestOptions] = None) -> BaseUser:
        """
        Get information about your user account, such as your username, email, and user ID.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.whoami()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/current-user/whoami", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def list(self, *, request_options: typing.Optional[RequestOptions] = None) -> typing.List[BaseUser]:
        """
        List all users in your Label Studio organization.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[BaseUser]


        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.list()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/users/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(typing.List[BaseUser], _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def create(
        self,
        *,
        id: typing.Optional[int] = OMIT,
        first_name: typing.Optional[str] = OMIT,
        last_name: typing.Optional[str] = OMIT,
        username: typing.Optional[str] = OMIT,
        email: typing.Optional[str] = OMIT,
        avatar: typing.Optional[str] = OMIT,
        initials: typing.Optional[str] = OMIT,
        phone: typing.Optional[str] = OMIT,
        allow_newsletters: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseUser:
        """
        Create a user in Label Studio.

        Parameters
        ----------
        id : typing.Optional[int]
            User ID

        first_name : typing.Optional[str]
            First name of the user

        last_name : typing.Optional[str]
            Last name of the user

        username : typing.Optional[str]
            Username of the user

        email : typing.Optional[str]
            Email of the user

        avatar : typing.Optional[str]
            Avatar URL of the user

        initials : typing.Optional[str]
            Initials of the user

        phone : typing.Optional[str]
            Phone number of the user

        allow_newsletters : typing.Optional[bool]
            Whether the user allows newsletters

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.create()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/users/",
            method="POST",
            json={
                "id": id,
                "first_name": first_name,
                "last_name": last_name,
                "username": username,
                "email": email,
                "avatar": avatar,
                "initials": initials,
                "phone": phone,
                "allow_newsletters": allow_newsletters,
            },
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def get(self, id: int, *, request_options: typing.Optional[RequestOptions] = None) -> BaseUser:
        """
        Get info about a specific Label Studio user.
        You will need to provide their user ID. You can find a list of all user IDs using [List users](list).

        Parameters
        ----------
        id : int
            User ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.get(
            id=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/users/{jsonable_encoder(id)}/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def delete(self, id: int, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete a specific Label Studio user.

        You will need to provide their user ID. You can find a list of all user IDs using [List users](list).

        <Warning>Use caution when deleting a user, as this can cause issues such as breaking the "Annotated by" filter or leaving orphaned records.</Warning>

        Parameters
        ----------
        id : int
            User ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.delete(
            id=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/users/{jsonable_encoder(id)}/", method="DELETE", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def update(
        self,
        id: int,
        *,
        users_update_request_id: typing.Optional[int] = OMIT,
        first_name: typing.Optional[str] = OMIT,
        last_name: typing.Optional[str] = OMIT,
        username: typing.Optional[str] = OMIT,
        email: typing.Optional[str] = OMIT,
        avatar: typing.Optional[str] = OMIT,
        initials: typing.Optional[str] = OMIT,
        phone: typing.Optional[str] = OMIT,
        allow_newsletters: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseUser:
        """
        Update details for a specific Label Studio user, such as their name or contact information.

        You will need to provide their user ID. You can find a list of all user IDs using [List users](list).

        Parameters
        ----------
        id : int
            User ID

        users_update_request_id : typing.Optional[int]
            User ID

        first_name : typing.Optional[str]
            First name of the user

        last_name : typing.Optional[str]
            Last name of the user

        username : typing.Optional[str]
            Username of the user

        email : typing.Optional[str]
            Email of the user

        avatar : typing.Optional[str]
            Avatar URL of the user

        initials : typing.Optional[str]
            Initials of the user

        phone : typing.Optional[str]
            Phone number of the user

        allow_newsletters : typing.Optional[bool]
            Whether the user allows newsletters

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseUser


        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.users.update(
            id=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/users/{jsonable_encoder(id)}/",
            method="PATCH",
            json={
                "id": users_update_request_id,
                "first_name": first_name,
                "last_name": last_name,
                "username": username,
                "email": email,
                "avatar": avatar,
                "initials": initials,
                "phone": phone,
                "allow_newsletters": allow_newsletters,
            },
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseUser, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pagination import AsyncPager, SyncPager
from ..core.pydantic_utilities import pydantic_v1
from ..core.request_options import RequestOptions
from ..types.base_task import BaseTask
from ..types.project_import import ProjectImport
from ..types.task import Task
from .types.tasks_list_request_fields import TasksListRequestFields
from .types.tasks_list_response import TasksListResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class TasksClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def create_many_status(
        self, id: int, import_pk: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectImport:
        """
        Get information about an async project import operation. This can be especially useful to monitor status, as large import jobs can take time.

        You will need the project ID and the unique ID of the import operation.

        The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list).

        The import ID is returned as part of the response when you call [Import tasks](import-tasks).

        Parameters
        ----------
        id : int
            The project ID.

        import_pk : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectImport


        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.create_many_status(
            id=1,
            import_pk="import_pk",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/projects/{jsonable_encoder(id)}/imports/{jsonable_encoder(import_pk)}/",
            method="GET",
            request_options=request_options,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(ProjectImport, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def delete_all_tasks(self, id: int, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete all tasks from a specific project.

        The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list).

        Parameters
        ----------
        id : int
            A unique integer value identifying this project.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.delete_all_tasks(
            id=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/projects/{jsonable_encoder(id)}/tasks/", method="DELETE", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def list(
        self,
        *,
        page: typing.Optional[int] = None,
        page_size: typing.Optional[int] = None,
        view: typing.Optional[int] = None,
        project: typing.Optional[int] = None,
        resolve_uri: typing.Optional[bool] = None,
        fields: typing.Optional[TasksListRequestFields] = None,
        review: typing.Optional[bool] = None,
        include: typing.Optional[str] = None,
        query: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SyncPager[Task]:
        """
        Retrieve a list of tasks.

        You can use the query parameters to filter the list by project and/or view (a tab within the Data Manager). You can also optionally add pagination to make the response easier to parse.

        The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list). The view ID can be found using [List views](../views/list).

        Parameters
        ----------
        page : typing.Optional[int]
            A page number within the paginated result set.

        page_size : typing.Optional[int]
            Number of results to return per page.

        view : typing.Optional[int]
            View ID

        project : typing.Optional[int]
            Project ID

        resolve_uri : typing.Optional[bool]
            Resolve task data URIs using Cloud Storage

        fields : typing.Optional[TasksListRequestFields]
            Set to "all" if you want to include annotations and predictions in the response

        review : typing.Optional[bool]
            Get tasks for review

        include : typing.Optional[str]
            Specify which fields to include in the response

        query : typing.Optional[str]
            Additional query to filter tasks. It must be JSON encoded string of dict containing one of the following parameters: `{"filters": ..., "selectedItems": ..., "ordering": ...}`. Check Data Manager > Create View for more details about filters, selectedItems and ordering.

            - **filters**: dict with `"conjunction"` string (`"or"` or `"and"`) and list of filters in `"items"` array. Each filter is a dictionary with keys: `"filter"`, `"operator"`, `"type"`, `"value"`. [Read more about available filters](https://labelstud.io/sdk/data_manager.html)<br/> Example: `{"conjunction": "or", "items": [{"filter": "filter:tasks:completed_at", "operator": "greater", "type": "Datetime", "value": "2021-01-01T00:00:00.000Z"}]}`
            - **selectedItems**: dictionary with keys: `"all"`, `"included"`, `"excluded"`. If "all" is `false`, `"included"` must be used. If "all" is `true`, `"excluded"` must be used.<br/> Examples: `{"all": false, "included": [1, 2, 3]}` or `{"all": true, "excluded": [4, 5]}`
            - **ordering**: list of fields to order by. Currently, ordering is supported by only one parameter. <br/>
              Example: `["completed_at"]`

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SyncPager[Task]
            List of Tasks

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.list()
        """
        page = page or 1
        _response = self._client_wrapper.httpx_client.request(
            "api/tasks/",
            method="GET",
            params={
                "page": page,
                "page_size": page_size,
                "view": view,
                "project": project,
                "resolve_uri": resolve_uri,
                "fields": fields,
                "review": review,
                "include": include,
                "query": query,
            },
            request_options=request_options,
        )
        if 200 <= _response.status_code < 300:
            _parsed_response = pydantic_v1.parse_obj_as(TasksListResponse, _response.json())  # type: ignore
            _has_next = True
            _get_next = lambda: self.list(
                page=page + 1,
                page_size=page_size,
                view=view,
                project=project,
                resolve_uri=resolve_uri,
                fields=fields,
                review=review,
                include=include,
                query=query,
                request_options=request_options,
            )
            _items = _parsed_response.tasks
            return SyncPager(has_next=_has_next, items=_items, get_next=_get_next)
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def create(
        self,
        *,
        data: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        project: typing.Optional[int] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseTask:
        """
        Create a new labeling task in Label Studio.

        The data you provide depends on your labeling config and data type.

        You will also need to provide a project ID. The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list).

        Parameters
        ----------
        data : typing.Optional[typing.Dict[str, typing.Any]]
            Task data dictionary with arbitrary keys and values

        project : typing.Optional[int]
            Project ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseTask
            Created task

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.create(
            data={"image": "https://example.com/image.jpg", "text": "Hello, world!"},
            project=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/tasks/",
            method="POST",
            json={"data": data, "project": project},
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseTask, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def get(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> BaseTask:
        """
        Get task data, metadata, annotations and other attributes for a specific labeling task by task ID.
        The task ID is available from the Label Studio URL when viewing the task, or you can retrieve it programmatically with [Get task list](list).

        Parameters
        ----------
        id : str
            Task ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseTask
            Task

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.get(
            id="id",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/tasks/{jsonable_encoder(id)}/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseTask, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def delete(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete a task in Label Studio.

        You will need the task ID. This is available from the Label Studio URL when viewing the task, or you can retrieve it programmatically with [Get task list](list).

        <Warning>This action cannot be undone.</Warning>

        Parameters
        ----------
        id : str
            Task ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.delete(
            id="id",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/tasks/{jsonable_encoder(id)}/", method="DELETE", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def update(
        self,
        id: str,
        *,
        data: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        project: typing.Optional[int] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseTask:
        """
        Update the attributes of an existing labeling task.

        You will need the task ID. This is available from the Label Studio URL when viewing the task, or you can retrieve it programmatically with [Get task list](list).

        Parameters
        ----------
        id : str
            Task ID

        data : typing.Optional[typing.Dict[str, typing.Any]]
            Task data dictionary with arbitrary keys and values

        project : typing.Optional[int]
            Project ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseTask
            Updated task

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.tasks.update(
            id="id",
            data={"image": "https://example.com/image.jpg", "text": "Hello, world!"},
            project=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"api/tasks/{jsonable_encoder(id)}/",
            method="PATCH",
            json={"data": data, "project": project},
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseTask, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncTasksClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def create_many_status(
        self, id: int, import_pk: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectImport:
        """
        Get information about an async project import operation. This can be especially useful to monitor status, as large import jobs can take time.

        You will need the project ID and the unique ID of the import operation.

        The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list).

        The import ID is returned as part of the response when you call [Import tasks](import-tasks).

        Parameters
        ----------
        id : int
            The project ID.

        import_pk : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectImport


        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.create_many_status(
            id=1,
            import_pk="import_pk",
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/projects/{jsonable_encoder(id)}/imports/{jsonable_encoder(import_pk)}/",
            method="GET",
            request_options=request_options,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(ProjectImport, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def delete_all_tasks(self, id: int, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete all tasks from a specific project.

        The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list).

        Parameters
        ----------
        id : int
            A unique integer value identifying this project.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.delete_all_tasks(
            id=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/projects/{jsonable_encoder(id)}/tasks/", method="DELETE", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def list(
        self,
        *,
        page: typing.Optional[int] = None,
        page_size: typing.Optional[int] = None,
        view: typing.Optional[int] = None,
        project: typing.Optional[int] = None,
        resolve_uri: typing.Optional[bool] = None,
        fields: typing.Optional[TasksListRequestFields] = None,
        review: typing.Optional[bool] = None,
        include: typing.Optional[str] = None,
        query: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncPager[Task]:
        """
        Retrieve a list of tasks.

        You can use the query parameters to filter the list by project and/or view (a tab within the Data Manager). You can also optionally add pagination to make the response easier to parse.

        The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list). The view ID can be found using [List views](../views/list).

        Parameters
        ----------
        page : typing.Optional[int]
            A page number within the paginated result set.

        page_size : typing.Optional[int]
            Number of results to return per page.

        view : typing.Optional[int]
            View ID

        project : typing.Optional[int]
            Project ID

        resolve_uri : typing.Optional[bool]
            Resolve task data URIs using Cloud Storage

        fields : typing.Optional[TasksListRequestFields]
            Set to "all" if you want to include annotations and predictions in the response

        review : typing.Optional[bool]
            Get tasks for review

        include : typing.Optional[str]
            Specify which fields to include in the response

        query : typing.Optional[str]
            Additional query to filter tasks. It must be JSON encoded string of dict containing one of the following parameters: `{"filters": ..., "selectedItems": ..., "ordering": ...}`. Check Data Manager > Create View for more details about filters, selectedItems and ordering.

            - **filters**: dict with `"conjunction"` string (`"or"` or `"and"`) and list of filters in `"items"` array. Each filter is a dictionary with keys: `"filter"`, `"operator"`, `"type"`, `"value"`. [Read more about available filters](https://labelstud.io/sdk/data_manager.html)<br/> Example: `{"conjunction": "or", "items": [{"filter": "filter:tasks:completed_at", "operator": "greater", "type": "Datetime", "value": "2021-01-01T00:00:00.000Z"}]}`
            - **selectedItems**: dictionary with keys: `"all"`, `"included"`, `"excluded"`. If "all" is `false`, `"included"` must be used. If "all" is `true`, `"excluded"` must be used.<br/> Examples: `{"all": false, "included": [1, 2, 3]}` or `{"all": true, "excluded": [4, 5]}`
            - **ordering**: list of fields to order by. Currently, ordering is supported by only one parameter. <br/>
              Example: `["completed_at"]`

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncPager[Task]
            List of Tasks

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.list()
        """
        page = page or 1
        _response = await self._client_wrapper.httpx_client.request(
            "api/tasks/",
            method="GET",
            params={
                "page": page,
                "page_size": page_size,
                "view": view,
                "project": project,
                "resolve_uri": resolve_uri,
                "fields": fields,
                "review": review,
                "include": include,
                "query": query,
            },
            request_options=request_options,
        )
        if 200 <= _response.status_code < 300:
            _parsed_response = pydantic_v1.parse_obj_as(TasksListResponse, _response.json())  # type: ignore
            _has_next = True
            _get_next = lambda: self.list(
                page=page + 1,
                page_size=page_size,
                view=view,
                project=project,
                resolve_uri=resolve_uri,
                fields=fields,
                review=review,
                include=include,
                query=query,
                request_options=request_options,
            )
            _items = _parsed_response.tasks
            return AsyncPager(has_next=_has_next, items=_items, get_next=_get_next)
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def create(
        self,
        *,
        data: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        project: typing.Optional[int] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseTask:
        """
        Create a new labeling task in Label Studio.

        The data you provide depends on your labeling config and data type.

        You will also need to provide a project ID. The project ID can be found in the URL when viewing the project in Label Studio, or you can retrieve all project IDs using [List all projects](../projects/list).

        Parameters
        ----------
        data : typing.Optional[typing.Dict[str, typing.Any]]
            Task data dictionary with arbitrary keys and values

        project : typing.Optional[int]
            Project ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseTask
            Created task

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.create(
            data={"image": "https://example.com/image.jpg", "text": "Hello, world!"},
            project=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/tasks/",
            method="POST",
            json={"data": data, "project": project},
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseTask, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def get(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> BaseTask:
        """
        Get task data, metadata, annotations and other attributes for a specific labeling task by task ID.
        The task ID is available from the Label Studio URL when viewing the task, or you can retrieve it programmatically with [Get task list](list).

        Parameters
        ----------
        id : str
            Task ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseTask
            Task

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.get(
            id="id",
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/tasks/{jsonable_encoder(id)}/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseTask, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def delete(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete a task in Label Studio.

        You will need the task ID. This is available from the Label Studio URL when viewing the task, or you can retrieve it programmatically with [Get task list](list).

        <Warning>This action cannot be undone.</Warning>

        Parameters
        ----------
        id : str
            Task ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.delete(
            id="id",
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/tasks/{jsonable_encoder(id)}/", method="DELETE", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def update(
        self,
        id: str,
        *,
        data: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        project: typing.Optional[int] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BaseTask:
        """
        Update the attributes of an existing labeling task.

        You will need the task ID. This is available from the Label Studio URL when viewing the task, or you can retrieve it programmatically with [Get task list](list).

        Parameters
        ----------
        id : str
            Task ID

        data : typing.Optional[typing.Dict[str, typing.Any]]
            Task data dictionary with arbitrary keys and values

        project : typing.Optional[int]
            Project ID

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BaseTask
            Updated task

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.tasks.update(
            id="id",
            data={"image": "https://example.com/image.jpg", "text": "Hello, world!"},
            project=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"api/tasks/{jsonable_encoder(id)}/",
            method="PATCH",
            json={"data": data, "project": project},
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(BaseTask, _response.json())  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from .types import (
    ActionsCreateRequestFilters,
    ActionsCreateRequestFiltersConjunction,
    ActionsCreateRequestFiltersItemsItem,
    ActionsCreateRequestFiltersItemsItemFilter,
    ActionsCreateRequestFiltersItemsItemOperator,
    ActionsCreateRequestId,
    ActionsCreateRequestOrderingItem,
    ActionsCreateRequestSelectedItems,
    ActionsCreateRequestSelectedItemsExcluded,
    ActionsCreateRequestSelectedItemsIncluded,
)

__all__ = [
    "ActionsCreateRequestFilters",
    "ActionsCreateRequestFiltersConjunction",
    "ActionsCreateRequestFiltersItemsItem",
    "ActionsCreateRequestFiltersItemsItemFilter",
    "ActionsCreateRequestFiltersItemsItemOperator",
    "ActionsCreateRequestId",
    "ActionsCreateRequestOrderingItem",
    "ActionsCreateRequestSelectedItems",
    "ActionsCreateRequestSelectedItemsExcluded",
    "ActionsCreateRequestSelectedItemsIncluded",
]



# This file was auto-generated by Fern from our API Definition.

import typing

AzureBlobImportStorageStatus = typing.Union[
    typing.Literal["initialized", "queued", "in_progress", "failed", "completed"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .annotation_last_action import AnnotationLastAction


class Annotation(pydantic_v1.BaseModel):
    id: typing.Optional[int] = None
    result: typing.Optional[typing.List[typing.Dict[str, typing.Any]]] = pydantic_v1.Field(default=None)
    """
    List of annotation results for the task
    """

    created_username: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Username string
    """

    created_ago: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Time delta from creation time
    """

    completed_by: typing.Optional[int] = None
    unique_id: typing.Optional[str] = None
    was_cancelled: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    User skipped the task
    """

    ground_truth: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    This annotation is a Ground Truth (ground_truth)
    """

    created_at: typing.Optional[dt.datetime] = pydantic_v1.Field(default=None)
    """
    Creation time
    """

    updated_at: typing.Optional[dt.datetime] = pydantic_v1.Field(default=None)
    """
    Last updated time
    """

    draft_created_at: typing.Optional[dt.datetime] = pydantic_v1.Field(default=None)
    """
    Draft creation time
    """

    lead_time: typing.Optional[float] = pydantic_v1.Field(default=None)
    """
    How much time it took to annotate the task
    """

    import_id: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Original annotation ID that was at the import step or NULL if this annotation wasn't imported
    """

    last_action: typing.Optional[AnnotationLastAction] = pydantic_v1.Field(default=None)
    """
    Action which was performed in the last annotation history item
    """

    task: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Corresponding task for this annotation
    """

    project: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Project ID for this annotation
    """

    updated_by: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Last user who updated this annotation
    """

    parent_prediction: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Points to the prediction from which this annotation was created
    """

    parent_annotation: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Points to the parent annotation from which this annotation was created
    """

    last_created_by: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    User who created the last annotation history item
    """

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        kwargs_with_defaults_exclude_none: typing.Any = {"by_alias": True, "exclude_none": True, **kwargs}

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset), super().dict(**kwargs_with_defaults_exclude_none)
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



import os
import json  # better to use "imports ujson as json" for the best performance

import uuid
import logging

from PIL import Image
from typing import Optional, Tuple
from urllib.request import (
    pathname2url,
)  # for converting "+","*", etc. in file paths to appropriate urls

from label_studio_sdk.converter.utils import ExpandFullPath
from label_studio_sdk.converter.imports.label_config import generate_label_config

logger = logging.getLogger("root")
default_image_root_url = "/data/local-files/?d=images"


def convert_yolo_to_ls(
    input_dir,
    out_file,
    to_name="image",
    from_name="label",
    out_type="annotations",
    image_root_url=default_image_root_url,
    image_ext=".jpg,.jpeg,.png",
    image_dims: Optional[Tuple[int, int]] = None,
):
    """Convert YOLO labeling to Label Studio JSON
    :param input_dir: directory with YOLO where images, labels, notes.json are located
    :param out_file: output file with Label Studio JSON tasks
    :param to_name: object name from Label Studio labeling config
    :param from_name: control tag name from Label Studio labeling config
    :param out_type: annotation type - "annotations" or "predictions"
    :param image_root_url: root URL path where images will be hosted, e.g.: http://example.com/images
    :param image_ext: image extension/s - single string or comma separated list to search, eg. .jpeg or .jpg, .png and so on.
    :param image_dims: image dimensions - optional tuple of integers specifying the image width and height of *all* images in the dataset. Defaults to opening the image to determine it's width and height, which is slower. This should only be used in the special case where you dataset has uniform image dimesions.
    """

    tasks = []
    logger.info("Reading YOLO notes and categories from %s", input_dir)

    # build categories=>labels dict
    notes_file = os.path.join(input_dir, "classes.txt")
    with open(notes_file) as f:
        lines = [line.strip() for line in f.readlines()]
    categories = {i: line for i, line in enumerate(lines)}
    logger.info(f"Found {len(categories)} categories")

    # generate and save labeling config
    label_config_file = out_file.replace(".json", "") + ".label_config.xml"
    generate_label_config(
        categories,
        {from_name: "RectangleLabels"},
        to_name,
        from_name,
        label_config_file,
    )

    # define directories
    labels_dir = os.path.join(input_dir, "labels")
    images_dir = os.path.join(input_dir, "images")
    logger.info("Converting labels from %s", labels_dir)

    # build array out of provided comma separated image_extns (str -> array)
    image_ext = [x.strip() for x in image_ext.split(",")]
    logger.info(f"image extensions->, {image_ext}")

    # loop through images
    for f in os.listdir(images_dir):
        image_file_found_flag = False
        for ext in image_ext:
            if f.endswith(ext):
                image_file = f
                image_file_base = os.path.splitext(f)[0]
                image_file_found_flag = True
                break
        if not image_file_found_flag:
            continue

        image_root_url += "" if image_root_url.endswith("/") else "/"
        task = {
            "data": {
                # eg. '../../foo+you.py' -> '../../foo%2Byou.py'
                "image": image_root_url
                + str(pathname2url(image_file))
            }
        }

        # define coresponding label file and check existence
        label_file = os.path.join(labels_dir, image_file_base + ".txt")

        if os.path.exists(label_file):
            task[out_type] = [
                {
                    "result": [],
                    "ground_truth": False,
                }
            ]

            # read image sizes
            if image_dims is None:
                # default to opening file if we aren't given image dims. slow!
                with Image.open(os.path.join(images_dir, image_file)) as im:
                    image_width, image_height = im.size
            else:
                image_width, image_height = image_dims

            with open(label_file) as file:
                # convert all bounding boxes to Label Studio Results
                lines = file.readlines()
                for line in lines:
                    values = line.split()
                    label_id, x, y, width, height = values[0:5]
                    score = float(values[5]) if len(values) >= 6 else None
                    x, y, width, height = (
                        float(x),
                        float(y),
                        float(width),
                        float(height),
                    )
                    item = {
                        "id": uuid.uuid4().hex[0:10],
                        "type": "rectanglelabels",
                        "value": {
                            "x": (x - width / 2) * 100,
                            "y": (y - height / 2) * 100,
                            "width": width * 100,
                            "height": height * 100,
                            "rotation": 0,
                            "rectanglelabels": [categories[int(label_id)]],
                        },
                        "to_name": to_name,
                        "from_name": from_name,
                        "image_rotation": 0,
                        "original_width": image_width,
                        "original_height": image_height,
                    }
                    if score:
                        item["score"] = score
                    task[out_type][0]["result"].append(item)

        tasks.append(task)

    if len(tasks) > 0:
        logger.info("Saving Label Studio JSON to %s", out_file)
        with open(out_file, "w") as out:
            json.dump(tasks, out)

        help_root_dir = ""
        if image_root_url == default_image_root_url:
            help_root_dir = (
                f"Set environment variables LABEL_STUDIO_LOCAL_FILES_SERVING_ENABLED=true and "
                f"LABEL_STUDIO_LOCAL_FILES_DOCUMENT_ROOT={input_dir} for Label Studio run,\n"
                f"add Local Storage with Absolute local path = {input_dir}/images"
            )

        print(
            "\n"
            f"  1. Create a new project in Label Studio\n"
            f'  2. Use Labeling Config from "{label_config_file}"\n'
            f"  3. Setup serving for images\n"
            f"       E.g. you can use Local Storage (or others):\n"
            f"       https://labelstud.io/guide/storage.html#Local-storage\n"
            f"       See tutorial here:\nhttps://github.com/HumanSignal/label-studio-converter/tree/master?tab=readme-ov-file#yolo-to-label-studio-converter\n"
            f"       {help_root_dir}\n"
            f'  4. Import "{out_file}" to the project\n'
        )
    else:
        logger.error("No labels converted")


def add_parser(subparsers):
    yolo = subparsers.add_parser("yolo")

    yolo.add_argument(
        "-i",
        "--input",
        dest="input",
        required=True,
        help="directory with YOLO where images, labels, notes.json are located",
        action=ExpandFullPath,
    )
    yolo.add_argument(
        "-o",
        "--output",
        dest="output",
        help="output file with Label Studio JSON tasks",
        default="output.json",
        action=ExpandFullPath,
    )
    yolo.add_argument(
        "--to-name",
        dest="to_name",
        help="object name from Label Studio labeling config",
        default="image",
    )
    yolo.add_argument(
        "--from-name",
        dest="from_name",
        help="control tag name from Label Studio labeling config",
        default="label",
    )
    yolo.add_argument(
        "--out-type",
        dest="out_type",
        help='annotation type - "annotations" or "predictions"',
        default="annotations",
    )
    yolo.add_argument(
        "--image-root-url",
        dest="image_root_url",
        help="root URL path where images will be hosted, e.g.: http://example.com/images",
        default=default_image_root_url,
    )
    yolo.add_argument(
        "--image-ext",
        dest="image_ext",
        help="image extension to search: .jpeg or .jpg, .png",
        default=".jpg",
    )
    yolo.add_argument(
        "--image-dims",
        dest="image_dims",
        type=int,
        nargs=2,
        help=(
            "optional tuple of integers specifying the image width and height of *all* "
            "images in the dataset. Defaults to opening the image to determine it's width "
            "and height, which is slower. This should only be used in the special "
            "case where you dataset has uniform image dimesions. e.g. `--image-dims 600 800` "
            "if all your images are of dimensions width=600, height=800"
        ),
        default=None,
    )



import os
import csv
import time
import logging
import ujson as json

from copy import deepcopy, copy

from label_studio_sdk.converter.utils import ensure_dir, get_annotator, prettify_result


logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")


def convert(item_iterator, input_data, output_dir, **kwargs):
    start_time = time.time()
    logger.debug("Convert CSV started")
    if str(output_dir).endswith(".csv"):
        output_file = output_dir
    else:
        ensure_dir(output_dir)
        output_file = os.path.join(output_dir, "result.csv")

    # these keys are always presented
    keys = {"annotator", "annotation_id", "created_at", "updated_at", "lead_time"}

    # make 2 passes: the first pass is to get keys, otherwise we can't write csv without headers
    logger.debug("Prepare column names for CSV ...")
    for item in item_iterator(input_data):
        record = prepare_annotation_keys(item)
        keys.update(record)

    # the second pass is to write records to csv
    logger.debug(
        f"Prepare done in {time.time()-start_time:0.2f} sec. Write CSV rows now ..."
    )
    with open(output_file, "w", encoding="utf8") as outfile:
        writer = csv.DictWriter(
            outfile,
            fieldnames=sorted(list(keys)),
            quoting=csv.QUOTE_NONNUMERIC,
            delimiter=kwargs["sep"],
        )
        writer.writeheader()

        for item in item_iterator(input_data):
            record = prepare_annotation(item)
            writer.writerow(record)

    logger.debug(f"CSV conversion finished in {time.time()-start_time:0.2f} sec")


def prepare_annotation(item):
    record = {}
    if item.get("id") is not None:
        record["id"] = item["id"]

    for name, value in item["output"].items():
        pretty_value = prettify_result(value)
        record[name] = (
            pretty_value
            if isinstance(pretty_value, str)
            else json.dumps(pretty_value, ensure_ascii=False)
        )

    for name, value in item["input"].items():
        if isinstance(value, dict) or isinstance(value, list):
            # flat dicts and arrays from task.data to json strings
            record[name] = json.dumps(value, ensure_ascii=False)
        else:
            record[name] = value

    record["annotator"] = get_annotator(item)
    record["annotation_id"] = item["annotation_id"]
    record["created_at"] = item["created_at"]
    record["updated_at"] = item["updated_at"]
    record["lead_time"] = item["lead_time"]

    if "agreement" in item:
        record["agreement"] = item["agreement"]

    if "history" in item and item["history"]:
        record["history"] = json.dumps(item["history"], ensure_ascii=False)

    return record


def prepare_annotation_keys(item):
    record = set(item["input"].keys())  # we don't need deepcopy for keys
    if item.get("id") is not None:
        record.add("id")

    for name, value in item["output"].items():
        record.add(name)

    if "agreement" in item:
        record.add("agreement")

    if "history" in item and item["history"]:
        record.add("history")

    return record



# This file was auto-generated by Fern from our API Definition.

from .types import (
    MlCreateRequestAuthMethod,
    MlCreateResponse,
    MlCreateResponseAuthMethod,
    MlUpdateRequestAuthMethod,
    MlUpdateResponse,
    MlUpdateResponseAuthMethod,
)

__all__ = [
    "MlCreateRequestAuthMethod",
    "MlCreateResponse",
    "MlCreateResponseAuthMethod",
    "MlUpdateRequestAuthMethod",
    "MlUpdateResponse",
    "MlUpdateResponseAuthMethod",
]



# This file was auto-generated by Fern from our API Definition.

import os
import typing

import httpx

from .actions.client import ActionsClient, AsyncActionsClient
from .annotations.client import AnnotationsClient, AsyncAnnotationsClient
from .core.api_error import ApiError
from .core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from .environment import LabelStudioEnvironment
from .export_storage.client import AsyncExportStorageClient, ExportStorageClient
from .files.client import AsyncFilesClient, FilesClient
from .import_storage.client import AsyncImportStorageClient, ImportStorageClient
from .ml.client import AsyncMlClient, MlClient
from .predictions.client import AsyncPredictionsClient, PredictionsClient
from .projects.client import AsyncProjectsClient, ProjectsClient
from .tasks.client import AsyncTasksClient, TasksClient
from .users.client import AsyncUsersClient, UsersClient
from .views.client import AsyncViewsClient, ViewsClient
from .webhooks.client import AsyncWebhooksClient, WebhooksClient


class LabelStudioBase:
    """
    Use this class to access the different functions within the SDK. You can instantiate any number of clients with different configuration that will propagate to these functions.

    Parameters
    ----------
    base_url : typing.Optional[str]
        The base url to use for requests from the client.

    environment : LabelStudioEnvironment
        The environment to use for requests from the client. from .environment import LabelStudioEnvironment



        Defaults to LabelStudioEnvironment.DEFAULT



    api_key : typing.Optional[str]
    timeout : typing.Optional[float]
        The timeout to be used, in seconds, for requests. By default the timeout is 60 seconds, unless a custom httpx client is used, in which case this default is not enforced.

    follow_redirects : typing.Optional[bool]
        Whether the default httpx client follows redirects or not, this is irrelevant if a custom httpx client is passed in.

    httpx_client : typing.Optional[httpx.Client]
        The httpx client to use for making requests, a preconfigured client is used by default, however this is useful should you want to pass in any custom httpx configuration.

    Examples
    --------
    from label_studio_sdk.client import LabelStudio

    client = LabelStudio(
        api_key="YOUR_API_KEY",
    )
    """

    def __init__(
        self,
        *,
        base_url: typing.Optional[str] = None,
        environment: LabelStudioEnvironment = LabelStudioEnvironment.DEFAULT,
        api_key: typing.Optional[str] = os.getenv("LABEL_STUDIO_API_KEY"),
        timeout: typing.Optional[float] = None,
        follow_redirects: typing.Optional[bool] = True,
        httpx_client: typing.Optional[httpx.Client] = None
    ):
        _defaulted_timeout = timeout if timeout is not None else 60 if httpx_client is None else None
        if api_key is None:
            raise ApiError(
                body="The client must be instantiated be either passing in api_key or setting LABEL_STUDIO_API_KEY"
            )
        self._client_wrapper = SyncClientWrapper(
            base_url=_get_base_url(base_url=base_url, environment=environment),
            api_key=api_key,
            httpx_client=httpx_client
            if httpx_client is not None
            else httpx.Client(timeout=_defaulted_timeout, follow_redirects=follow_redirects)
            if follow_redirects is not None
            else httpx.Client(timeout=_defaulted_timeout),
            timeout=_defaulted_timeout,
        )
        self.annotations = AnnotationsClient(client_wrapper=self._client_wrapper)
        self.users = UsersClient(client_wrapper=self._client_wrapper)
        self.actions = ActionsClient(client_wrapper=self._client_wrapper)
        self.views = ViewsClient(client_wrapper=self._client_wrapper)
        self.files = FilesClient(client_wrapper=self._client_wrapper)
        self.projects = ProjectsClient(client_wrapper=self._client_wrapper)
        self.ml = MlClient(client_wrapper=self._client_wrapper)
        self.predictions = PredictionsClient(client_wrapper=self._client_wrapper)
        self.tasks = TasksClient(client_wrapper=self._client_wrapper)
        self.import_storage = ImportStorageClient(client_wrapper=self._client_wrapper)
        self.export_storage = ExportStorageClient(client_wrapper=self._client_wrapper)
        self.webhooks = WebhooksClient(client_wrapper=self._client_wrapper)


class AsyncLabelStudioBase:
    """
    Use this class to access the different functions within the SDK. You can instantiate any number of clients with different configuration that will propagate to these functions.

    Parameters
    ----------
    base_url : typing.Optional[str]
        The base url to use for requests from the client.

    environment : LabelStudioEnvironment
        The environment to use for requests from the client. from .environment import LabelStudioEnvironment



        Defaults to LabelStudioEnvironment.DEFAULT



    api_key : typing.Optional[str]
    timeout : typing.Optional[float]
        The timeout to be used, in seconds, for requests. By default the timeout is 60 seconds, unless a custom httpx client is used, in which case this default is not enforced.

    follow_redirects : typing.Optional[bool]
        Whether the default httpx client follows redirects or not, this is irrelevant if a custom httpx client is passed in.

    httpx_client : typing.Optional[httpx.AsyncClient]
        The httpx client to use for making requests, a preconfigured client is used by default, however this is useful should you want to pass in any custom httpx configuration.

    Examples
    --------
    from label_studio_sdk.client import AsyncLabelStudio

    client = AsyncLabelStudio(
        api_key="YOUR_API_KEY",
    )
    """

    def __init__(
        self,
        *,
        base_url: typing.Optional[str] = None,
        environment: LabelStudioEnvironment = LabelStudioEnvironment.DEFAULT,
        api_key: typing.Optional[str] = os.getenv("LABEL_STUDIO_API_KEY"),
        timeout: typing.Optional[float] = None,
        follow_redirects: typing.Optional[bool] = True,
        httpx_client: typing.Optional[httpx.AsyncClient] = None
    ):
        _defaulted_timeout = timeout if timeout is not None else 60 if httpx_client is None else None
        if api_key is None:
            raise ApiError(
                body="The client must be instantiated be either passing in api_key or setting LABEL_STUDIO_API_KEY"
            )
        self._client_wrapper = AsyncClientWrapper(
            base_url=_get_base_url(base_url=base_url, environment=environment),
            api_key=api_key,
            httpx_client=httpx_client
            if httpx_client is not None
            else httpx.AsyncClient(timeout=_defaulted_timeout, follow_redirects=follow_redirects)
            if follow_redirects is not None
            else httpx.AsyncClient(timeout=_defaulted_timeout),
            timeout=_defaulted_timeout,
        )
        self.annotations = AsyncAnnotationsClient(client_wrapper=self._client_wrapper)
        self.users = AsyncUsersClient(client_wrapper=self._client_wrapper)
        self.actions = AsyncActionsClient(client_wrapper=self._client_wrapper)
        self.views = AsyncViewsClient(client_wrapper=self._client_wrapper)
        self.files = AsyncFilesClient(client_wrapper=self._client_wrapper)
        self.projects = AsyncProjectsClient(client_wrapper=self._client_wrapper)
        self.ml = AsyncMlClient(client_wrapper=self._client_wrapper)
        self.predictions = AsyncPredictionsClient(client_wrapper=self._client_wrapper)
        self.tasks = AsyncTasksClient(client_wrapper=self._client_wrapper)
        self.import_storage = AsyncImportStorageClient(client_wrapper=self._client_wrapper)
        self.export_storage = AsyncExportStorageClient(client_wrapper=self._client_wrapper)
        self.webhooks = AsyncWebhooksClient(client_wrapper=self._client_wrapper)


def _get_base_url(*, base_url: typing.Optional[str] = None, environment: LabelStudioEnvironment) -> str:
    if base_url is not None:
        return base_url
    elif environment is not None:
        return environment.value
    else:
        raise Exception("Please pass in either base_url or environment to construct the client")



# This file was auto-generated by Fern from our API Definition.

from .export_storage_list_types_response_item import ExportStorageListTypesResponseItem

__all__ = ["ExportStorageListTypesResponseItem"]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .project_sampling import ProjectSampling
from .project_skip_queue import ProjectSkipQueue
from .user_simple import UserSimple


class Project(pydantic_v1.BaseModel):
    id: typing.Optional[int] = None
    title: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Project name. Must be between 3 and 50 characters long.
    """

    description: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Project description
    """

    label_config: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Label config in XML format. See more about it in documentation
    """

    expert_instruction: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Labeling instructions in HTML format
    """

    show_instruction: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Show instructions to the annotator before they start
    """

    show_skip_button: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Show a skip button in interface and allow annotators to skip the task
    """

    enable_empty_annotation: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Allow annotators to submit empty annotations
    """

    show_annotation_history: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Show annotation history to annotator
    """

    organization: typing.Optional[int] = None
    color: typing.Optional[str] = None
    maximum_annotations: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Maximum number of annotations for one task. If the number of annotations per task is equal or greater to this value, the task is completed (is_labeled=True)
    """

    is_published: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Whether or not the project is published to annotators
    """

    model_version: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Machine learning model version
    """

    is_draft: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Whether or not the project is in the middle of being created
    """

    created_by: typing.Optional[UserSimple] = None
    created_at: typing.Optional[dt.datetime] = None
    min_annotations_to_start_training: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Minimum number of completed tasks after which model training is started
    """

    start_training_on_annotation_update: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Start model training after any annotations are submitted or updated
    """

    show_collab_predictions: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    If set, the annotator can view model predictions
    """

    num_tasks_with_annotations: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Tasks with annotations count
    """

    task_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Total task number in project
    """

    useful_annotation_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Useful annotation number in project not including skipped_annotations_number and ground_truth_number. Total annotations = annotation_number + skipped_annotations_number + ground_truth_number
    """

    ground_truth_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Honeypot annotation number in project
    """

    skipped_annotations_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Skipped by collaborators annotation number in project
    """

    total_annotations_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Total annotations number in project including skipped_annotations_number and ground_truth_number.
    """

    total_predictions_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Total predictions number in project including skipped_annotations_number, ground_truth_number, and useful_annotation_number.
    """

    sampling: typing.Optional[ProjectSampling] = None
    show_ground_truth_first: typing.Optional[bool] = None
    show_overlap_first: typing.Optional[bool] = None
    overlap_cohort_percentage: typing.Optional[int] = None
    task_data_login: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Task data credentials: login
    """

    task_data_password: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Task data credentials: password
    """

    control_weights: typing.Optional[typing.Dict[str, typing.Any]] = pydantic_v1.Field(default=None)
    """
    Dict of weights for each control tag in metric calculation. Each control tag (e.g. label or choice) will have it's own key in control weight dict with weight for each label and overall weight.For example, if bounding box annotation with control tag named my_bbox should be included with 0.33 weight in agreement calculation, and the first label Car should be twice more important than Airplaine, then you have to need the specify: {'my_bbox': {'type': 'RectangleLabels', 'labels': {'Car': 1.0, 'Airplaine': 0.5}, 'overall': 0.33}
    """

    parsed_label_config: typing.Optional[typing.Dict[str, typing.Any]] = pydantic_v1.Field(default=None)
    """
    JSON-formatted labeling configuration
    """

    evaluate_predictions_automatically: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Retrieve and display predictions when loading a task
    """

    config_has_control_tags: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Flag to detect is project ready for labeling
    """

    skip_queue: typing.Optional[ProjectSkipQueue] = None
    reveal_preannotations_interactively: typing.Optional[bool] = pydantic_v1.Field(default=None)
    """
    Reveal pre-annotations interactively
    """

    pinned_at: typing.Optional[dt.datetime] = pydantic_v1.Field(default=None)
    """
    Pinned date and time
    """

    finished_task_number: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Finished tasks
    """

    queue_total: typing.Optional[str] = None
    queue_done: typing.Optional[str] = None

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        kwargs_with_defaults_exclude_none: typing.Any = {"by_alias": True, "exclude_none": True, **kwargs}

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset), super().dict(**kwargs_with_defaults_exclude_none)
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



from pydantic import BaseModel
from typing import Dict, Optional, List, Tuple, Any, Callable


def get_tag_class(name, default_mapping, re_mapping=None):
    """ """
    lc_name = name.lower()
    if re_mapping and lc_name in re_mapping:
        return re_mapping.get(lc_name)
    else:
        class_name = default_mapping.get(name.lower())
        return class_name


class LabelStudioTag(BaseModel):
    """
    Base class for a LabelStudio Tag

    Attributes:
    -----------
    attr: Optional[Dict]
        A dictionary of attributes for the tag
    tag: Optional[str]
        The tag name
    """

    attr: Optional[Dict]
    tag: Optional[str]

    def match(
        self,
        tag_type: str,
        name: Optional[str] = None,
        name_filter_fn: Optional[Callable] = None,
        to_name: Optional[str] = None,
        to_name_filter_fn: Optional[Callable] = None,
    ) -> bool:
        """
        This method checks if the current instance of LabelStudioTag matches the provided parameters.

        Parameters:
        -----------
        tag_type : str
            The type of the tag to match. It can be a string or a tuple of strings.
        name : Optional[str]
            The name of the tag to match. If provided, it should match the name of the current instance.
        name_filter_fn : Optional[Callable]
            A function to filter the name of the tag. If provided, the function should return True for a match.
        to_name : Optional[str]
            The 'to_name' attribute of the tag to match. If provided, it should match the 'to_name' of the current instance.
        to_name_filter_fn : Optional[Callable]
            A function to filter the 'to_name' of the tag. If provided, the function should return True for a match.

        Returns:
        --------
        bool
            True if the current instance matches the provided parameters, False otherwise.
        """
        if isinstance(tag_type, str):
            tag_type = tag_type.lower()
        elif isinstance(tag_type, tuple):
            tag_type = tuple(t.lower() for t in tag_type)

        if name:
            name = name.lower()

        if to_name:
            to_name = to_name.lower()

        if (isinstance(tag_type, str) and self.tag.lower() != tag_type) or (
            isinstance(tag_type, tuple) and self.tag.lower() not in tag_type
        ):
            return False

        if name and self.name.lower() != name:
            return False

        if name_filter_fn and not name_filter_fn(self.name.lower()):
            return False

        if to_name and self.to_name.lower() != to_name:
            return False

        if to_name_filter_fn and not to_name_filter_fn(self.name.lower()):
            return False

        return True



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1


class Prediction(pydantic_v1.BaseModel):
    id: typing.Optional[int] = None
    result: typing.List[typing.Dict[str, typing.Any]] = pydantic_v1.Field()
    """
    List of prediction results for the task
    """

    model_version: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Model version - tag for predictions that can be used to filter tasks in Data Manager, as well as select specific model version for showing preannotations in the labeling interface
    """

    created_ago: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    Delta time from creation time
    """

    score: typing.Optional[float] = pydantic_v1.Field(default=None)
    """
    Prediction score
    """

    cluster: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    Cluster for the current prediction
    """

    neighbors: typing.Optional[typing.Dict[str, typing.Any]] = pydantic_v1.Field(default=None)
    """
    Array of task IDs of the closest neighbors
    """

    mislabeling: typing.Optional[float] = pydantic_v1.Field(default=None)
    """
    Related task mislabeling score
    """

    created_at: typing.Optional[dt.datetime] = None
    updated_at: typing.Optional[dt.datetime] = None
    model: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    An ML Backend instance that created the prediction.
    """

    model_run: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    A run of a ModelVersion that created the prediction.
    """

    task: int
    project: typing.Optional[int] = None

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        kwargs_with_defaults_exclude_none: typing.Any = {"by_alias": True, "exclude_none": True, **kwargs}

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset), super().dict(**kwargs_with_defaults_exclude_none)
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



""" # Data Manager module for the Label Studio SDK

    Classes can be used to filter, order, and select items in `label_studio_sdk.project.Project.get_tasks`
    and provides enumeration for all column names available in the Data Manager for tasks, and other helpers.

    See the [client](client.html), [project](project.html) or [utils](utils.html) modules for other operations you
    might want to perform.

    Example:

    ```python
    from label_studio_sdk.data_manager import Filters, Column, Operator, Type

    filters = Filters.create(Filters.OR, [
        Filters.item(
            Column.id,
            Operator.GREATER,
            Type.Number,
            Filters.value(42)
        ),
        Filters.item(
            Column.completed_at,
            Operator.IN,
            Type.Datetime,
            Filters.value(
                datetime(2021, 11, 1),
                datetime.now()
            )
        )
    ])
    tasks = project.get_tasks(filters=filters)
    ```
"""

from datetime import datetime

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class Filters:
    """
    Use the methods and variables in this class to create and combine filters for tasks on the Label Studio Data Manager.
    """

    OR = "or"
    """Combine filters with an OR"""
    AND = "and"
    """Combine filters with an AND"""

    @staticmethod
    def create(conjunction, items):
        """Create a filter for `label_studio_sdk.project.Project.get_tasks()`

        Parameters
        ----------
        conjunction: str
            The conjunction operator between filters ('or' or 'and')
        items: list
            What to filter, use `Filter.item()` method to build it

        Returns
        -------
        dict
            containing specified parameters

        """
        return {"conjunction": conjunction, "items": items}

    @staticmethod
    def item(name, operator, column_type, value):
        """Use in combination with other classes to specify the contents of a filter.

        Parameters
        ----------
        name: `Column` or str
            Column.id, Column.completed_at, Column.data('my_field'), etc
        operator: `Operator`
            Operator.EQUAL, Operator.GREATER_OR_EQUAL, Operator.IN, etc
        column_type: `Type`
            Type.Number, Type.Boolean, Type.String, etc
        value: `Filters.value()`
            Filters.value(42), Filters.value('test'), Filters.value(datetime(2021, 01, 01), datetime.now())

        Returns
        -------
        dict
        """
        return {
            "filter": "filter:" + name,
            "operator": operator,
            "type": column_type,
            "value": value,
        }

    @staticmethod
    def datetime(dt):
        """Date time string format for filtering the Data Manager.

        Parameters
        ----------
        dt
            datetime instance

        Returns
        -------
        str
            datetime in `'%Y-%m-%dT%H:%M:%S.%fZ'` format

        """
        assert isinstance(dt, datetime), "dt must be datetime type"
        return dt.strftime(DATETIME_FORMAT)

    @classmethod
    def value(cls, value, maximum=None):
        """Set a filter value in the Data Manager.

        Parameters
        ----------
        value: str | int | float | datetime | boolean
            value to use for filtering. If the maximum parameter is passed, then this value field is the minimum.

        maximum: int | float  | datetime
            Specify a maximum for a filtering range with IN, NOT_IN operators.

        Returns
        -------
        any
            value for filtering

        """
        if isinstance(value, datetime):
            value = cls.datetime(value)

        if maximum is not None:
            if isinstance(maximum, datetime):
                maximum = cls.datetime(maximum)
            return {"min": value, "max": maximum}

        return value


class Operator:
    """Specify the operator to use when creating a filter."""

    EQUAL = "equal"
    NOT_EQUAL = "not_equal"
    LESS = "less"
    GREATER = "greater"
    LESS_OR_EQUAL = "less_or_equal"
    GREATER_OR_EQUAL = "greater_or_equal"
    IN = "in"
    NOT_IN = "not_in"
    IN_LIST = "in_list"
    NOT_IN_LIST = "not_in_list"
    EMPTY = "empty"
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    REGEX = "regex"


class Type:
    """Specify the type of data in a column."""

    Number = "Number"
    Datetime = "Datetime"
    Boolean = "Boolean"
    String = "String"
    List = "List"

    Unknown = "Unknown"
    """ Unknown is explicitly converted to string format. """


class Column:
    """Specify the column on the Data Manager in Label Studio UI to use in the filter."""

    id = "tasks:id"
    """Task ID"""
    inner_id = "tasks:inner_id"
    """Task Inner ID, it starts from 1 for all projects"""
    ground_truth = "tasks:ground_truth"
    """Ground truth status of the tasks"""
    annotations_results = "tasks:annotations_results"
    """Annotation results for the tasks"""
    reviewed = "tasks:reviewed"
    """Whether the tasks have been reviewed (Enterprise only)"""
    predictions_score = "tasks:predictions_score"
    """Prediction score for the task"""
    predictions_model_versions = "tasks:predictions_model_versions"
    """Model version used for the predictions"""
    predictions_results = "tasks:predictions_results"
    """Prediction results for the tasks"""
    file_upload = "tasks:file_upload"
    """Name of the file uploaded to create the tasks"""
    created_at = "tasks:created_at"
    """Time the task was created at"""
    updated_at = "tasks:updated_at"
    """Time the task was updated at (e.g. new annotation was created, review added, etc)"""
    annotators = "tasks:annotators"
    """Annotators that completed the task (Community). Can include assigned annotators (Enterprise only)"""
    total_predictions = "tasks:total_predictions"
    """Total number of predictions for the task"""
    cancelled_annotations = "tasks:cancelled_annotations"
    """Number of cancelled or skipped annotations for the task"""
    total_annotations = "tasks:total_annotations"
    """Total number of annotations on a task"""
    completed_at = "tasks:completed_at"
    """Time when a task was fully annotated"""
    agreement = "tasks:agreement"
    """Agreement for annotation results for a specific task (Enterprise only)"""
    reviewers = "tasks:reviewers"
    """Reviewers that reviewed the task, or assigned reviewers (Enterprise only)"""
    reviews_rejected = "tasks:reviews_rejected"
    """Number of annotations rejected for a task in review (Enterprise only)"""
    reviews_accepted = "tasks:reviews_accepted"
    """Number of annotations accepted for a task in review (Enterprise only)"""
    comments = "tasks:comments"
    """Number of comments in a task"""
    unresolved_comment_count = "tasks:unresolved_comment_count"
    """Number of unresolved comments in a task"""

    @staticmethod
    def data(task_field):
        """Create a filter name for the task data field

        Parameters
        ----------
        task_field

        Returns
        -------
        str
            Filter name for task data

        """
        return "tasks:data." + task_field


def _test():
    """Test it"""
    filters = Filters.create(
        Filters.OR,
        [
            Filters.item(Column.id, Operator.GREATER, Type.Number, Filters.value(42)),
            Filters.item(
                Column.completed_at,
                Operator.IN,
                Type.Datetime,
                Filters.value(
                    datetime(2021, 11, 1),
                    datetime(2021, 11, 5),
                ),
            ),
        ],
    )

    assert filters == {
        "conjunction": "or",
        "items": [
            {
                "filter": "filter:tasks:id",
                "operator": "greater",
                "type": "Number",
                "value": 42,
            },
            {
                "filter": "filter:tasks:completed_at",
                "operator": "in",
                "type": "Datetime",
                "value": {
                    "min": "2021-11-01T00:00:00.000000Z",
                    "max": "2021-11-05T00:00:00.000000Z",
                },
            },
        ],
    }



from label_studio_sdk.converter.imports.colors import COLORS


LABELS = """
  <{# TAG_NAME #} name="{# FROM_NAME #}" toName="image">
{# LABELS #}  </{# TAG_NAME #}>
"""

LABELING_CONFIG = """<View>
  <Image name="{# TO_NAME #}" value="$image"/>
{# BODY #}</View>
"""


def generate_label_config(
    categories, tags, to_name="image", from_name="label", filename=None
):
    labels = ""
    for key in sorted(categories.keys()):
        color = COLORS[int(key) % len(COLORS)]
        label = f'    <Label value="{categories[key]}" background="rgba({color[0]}, {color[1]}, {color[2]}, 1)"/>\n'
        labels += label

    body = ""
    for from_name in tags:
        tag_body = (
            str(LABELS)
            .replace("{# TAG_NAME #}", tags[from_name])
            .replace("{# LABELS #}", labels)
            .replace("{# TO_NAME #}", to_name)
            .replace("{# FROM_NAME #}", from_name)
        )
        body += f'\n  <Header value="{tags[from_name]}"/>' + tag_body

    config = (
        str(LABELING_CONFIG)
        .replace("{# BODY #}", body)
        .replace("{# TO_NAME #}", to_name)
    )

    if filename:
        with open(filename, "w") as f:
            f.write(config)

    return config



# This file was auto-generated by Fern from our API Definition.

import typing

ViewsUpdateRequestDataFiltersItemsItemOperator = typing.Union[
    typing.Literal[
        "contains",
        "ends_with",
        "equal",
        "exists",
        "greater",
        "greater_or_equal",
        "in",
        "less",
        "less_or_equal",
        "not_contains",
        "not_equal",
        "not_exists",
        "not_in",
        "starts_with",
    ],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ...core.datetime_utils import serialize_datetime
from ...core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .actions_create_request_filters_items_item_filter import ActionsCreateRequestFiltersItemsItemFilter
from .actions_create_request_filters_items_item_operator import ActionsCreateRequestFiltersItemsItemOperator


class ActionsCreateRequestFiltersItemsItem(pydantic_v1.BaseModel):
    filter: ActionsCreateRequestFiltersItemsItemFilter = pydantic_v1.Field()
    """
    Filter identifier, it should start with `filter:tasks:` prefix, e.g. `filter:tasks:agreement`. For `task.data` fields it may look like `filter:tasks:data.field_name`. Possible values:<br><li>`filter:tasks:agreement`<br> (Number) Agreement for annotation results for a specific task (Enterprise only)</li><br><li>`filter:tasks:annotations_results`<br> (String) Annotation results for the tasks</li><br><li>`filter:tasks:annotators`<br> (List) Annotators that completed the task (Community). Can include assigned annotators (Enterprise only). Important note: the filter `type` should be List, but the filter `value` is integer</li><br><li>`filter:tasks:cancelled_annotations`<br> (Number) Number of cancelled or skipped annotations for the task</li><br><li>`filter:tasks:comments`<br> (Number) Number of comments in a task</li><br><li>`filter:tasks:completed_at`<br> (Datetime) Time when a task was fully annotated</li><br><li>`filter:tasks:created_at`<br> (Datetime) Time the task was created at</li><br><li>`filter:tasks:file_upload`<br> (String) Name of the file uploaded to create the tasks</li><br><li>`filter:tasks:ground_truth`<br> (Boolean) Ground truth status of the tasks</li><br><li>`filter:tasks:id`<br> (Number) Task ID</li><br><li>`filter:tasks:inner_id`<br> (Number) Task Inner ID, it starts from 1 for all projects</li><br><li>`filter:tasks:predictions_model_versions`<br> (String) Model version used for the predictions</li><br><li>`filter:tasks:predictions_results`<br> (String) Prediction results for the tasks</li><br><li>`filter:tasks:predictions_score`<br> (Number) Prediction score for the task</li><br><li>`filter:tasks:reviewed`<br> (Boolean) Whether the tasks have been reviewed (Enterprise only)</li><br><li>`filter:tasks:reviewers`<br> (String) Reviewers that reviewed the task, or assigned reviewers (Enterprise only). Important note: the filter `type` should be List, but the filter `value` is integer</li><br><li>`filter:tasks:reviews_accepted`<br> (Number) Number of annotations accepted for a task in review (Enterprise only)</li><br><li>`filter:tasks:reviews_rejected`<br> (Number) Number of annotations rejected for a task in review (Enterprise only)</li><br><li>`filter:tasks:total_annotations`<br> (Number) Total number of annotations on a task</li><br><li>`filter:tasks:total_predictions`<br> (Number) Total number of predictions for the task</li><br><li>`filter:tasks:unresolved_comment_count`<br> (Number) Number of unresolved comments in a task</li><br><li>`filter:tasks:updated_at`<br> (Datetime) Time the task was updated at (e.g. new annotation was created, review added, etc)</li>
    """

    operator: ActionsCreateRequestFiltersItemsItemOperator = pydantic_v1.Field()
    """
    Filter operator. Possible values:<br><li>`contains`<br> Contains</li><br><li>`ends_with`<br> Ends with</li><br><li>`equal`<br> Equal to</li><br><li>`exists`<br> Exists</li><br><li>`greater`<br> Greater than</li><br><li>`greater_or_equal`<br> Greater than or equal to</li><br><li>`in`<br> Is between min and max values, so the filter `value` should be e.g. `{"min": 1, "max": 7}`</li><br><li>`less`<br> Less than</li><br><li>`less_or_equal`<br> Less than or equal to</li><br><li>`not_contains`<br> Does not contain</li><br><li>`not_equal`<br> Not equal to</li><br><li>`not_exists`<br> Does not exist</li><br><li>`not_in`<br> Is not between min and max values, so the filter `value` should be e.g. `{"min": 1, "max": 7}`</li><br><li>`starts_with`<br> Starts with</li>
    """

    type: str = pydantic_v1.Field()
    """
    Type of the filter value. Possible values:<br><li>`Boolean`<br> Boolean</li><br><li>`Datetime`<br> Datetime string in `strftime('%Y-%m-%dT%H:%M:%S.%fZ')` format</li><br><li>`List`<br> List of items</li><br><li>`Number`<br> Float or Integer</li><br><li>`String`<br> String</li><br><li>`Unknown`<br> Unknown is explicitly converted to string format</li>
    """

    value: str = pydantic_v1.Field()
    """
    Value to filter by
    """

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        kwargs_with_defaults_exclude_none: typing.Any = {"by_alias": True, "exclude_none": True, **kwargs}

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset), super().dict(**kwargs_with_defaults_exclude_none)
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



# This file was auto-generated by Fern from our API Definition.

from .local_create_response import LocalCreateResponse
from .local_update_response import LocalUpdateResponse

__all__ = ["LocalCreateResponse", "LocalUpdateResponse"]



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .types.actions_create_request_filters import ActionsCreateRequestFilters
from .types.actions_create_request_id import ActionsCreateRequestId
from .types.actions_create_request_ordering_item import ActionsCreateRequestOrderingItem
from .types.actions_create_request_selected_items import ActionsCreateRequestSelectedItems

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ActionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list(self, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Retrieve all the registered actions with descriptions that data manager can use.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.actions.list()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/dm/actions/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def create(
        self,
        *,
        id: typing.Optional[ActionsCreateRequestId] = None,
        project: typing.Optional[int] = None,
        view: typing.Optional[int] = None,
        filters: typing.Optional[ActionsCreateRequestFilters] = OMIT,
        selected_items: typing.Optional[ActionsCreateRequestSelectedItems] = OMIT,
        ordering: typing.Optional[typing.Sequence[ActionsCreateRequestOrderingItem]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Perform a Data Manager action with the selected tasks and filters.

        Parameters
        ----------
        id : typing.Optional[ActionsCreateRequestId]
            Action name ID, the full list of actions can be retrieved with a GET request

        project : typing.Optional[int]
            Project ID

        view : typing.Optional[int]
            View ID (optional, it has higher priority than filters, selectedItems and ordering from the request body payload)

        filters : typing.Optional[ActionsCreateRequestFilters]
            Filters to apply on tasks. You can use [the helper class `Filters` from this page](https://labelstud.io/sdk/data_manager.html) to create Data Manager Filters.<br>Example: `{"conjunction": "or", "items": [{"filter": "filter:tasks:completed_at", "operator": "greater", "type": "Datetime", "value": "2021-01-01T00:00:00.000Z"}]}`

        selected_items : typing.Optional[ActionsCreateRequestSelectedItems]
            Task selection by IDs. If filters are applied, the selection will be applied to the filtered tasks.If "all" is `false`, `"included"` must be used. If "all" is `true`, `"excluded"` must be used.<br>Examples: `{"all": false, "included": [1, 2, 3]}` or `{"all": true, "excluded": [4, 5]}`

        ordering : typing.Optional[typing.Sequence[ActionsCreateRequestOrderingItem]]
            List of fields to order by. Fields are similar to filters but without the `filter:` prefix. To reverse the order, add a minus sign before the field name, e.g. `-tasks:created_at`.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import LabelStudio

        client = LabelStudio(
            api_key="YOUR_API_KEY",
        )
        client.actions.create()
        """
        _response = self._client_wrapper.httpx_client.request(
            "api/dm/actions/",
            method="POST",
            params={"id": id, "project": project, "view": view},
            json={"filters": filters, "selectedItems": selected_items, "ordering": ordering},
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncActionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list(self, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Retrieve all the registered actions with descriptions that data manager can use.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.actions.list()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/dm/actions/", method="GET", request_options=request_options
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def create(
        self,
        *,
        id: typing.Optional[ActionsCreateRequestId] = None,
        project: typing.Optional[int] = None,
        view: typing.Optional[int] = None,
        filters: typing.Optional[ActionsCreateRequestFilters] = OMIT,
        selected_items: typing.Optional[ActionsCreateRequestSelectedItems] = OMIT,
        ordering: typing.Optional[typing.Sequence[ActionsCreateRequestOrderingItem]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Perform a Data Manager action with the selected tasks and filters.

        Parameters
        ----------
        id : typing.Optional[ActionsCreateRequestId]
            Action name ID, the full list of actions can be retrieved with a GET request

        project : typing.Optional[int]
            Project ID

        view : typing.Optional[int]
            View ID (optional, it has higher priority than filters, selectedItems and ordering from the request body payload)

        filters : typing.Optional[ActionsCreateRequestFilters]
            Filters to apply on tasks. You can use [the helper class `Filters` from this page](https://labelstud.io/sdk/data_manager.html) to create Data Manager Filters.<br>Example: `{"conjunction": "or", "items": [{"filter": "filter:tasks:completed_at", "operator": "greater", "type": "Datetime", "value": "2021-01-01T00:00:00.000Z"}]}`

        selected_items : typing.Optional[ActionsCreateRequestSelectedItems]
            Task selection by IDs. If filters are applied, the selection will be applied to the filtered tasks.If "all" is `false`, `"included"` must be used. If "all" is `true`, `"excluded"` must be used.<br>Examples: `{"all": false, "included": [1, 2, 3]}` or `{"all": true, "excluded": [4, 5]}`

        ordering : typing.Optional[typing.Sequence[ActionsCreateRequestOrderingItem]]
            List of fields to order by. Fields are similar to filters but without the `filter:` prefix. To reverse the order, add a minus sign before the field name, e.g. `-tasks:created_at`.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from label_studio_sdk.client import AsyncLabelStudio

        client = AsyncLabelStudio(
            api_key="YOUR_API_KEY",
        )
        await client.actions.create()
        """
        _response = await self._client_wrapper.httpx_client.request(
            "api/dm/actions/",
            method="POST",
            params={"id": id, "project": project, "view": view},
            json={"filters": filters, "selectedItems": selected_items, "ordering": ordering},
            request_options=request_options,
            omit=OMIT,
        )
        if 200 <= _response.status_code < 300:
            return
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from .types import ProjectsCreateResponse, ProjectsImportTasksResponse, ProjectsListResponse
from . import exports

__all__ = ["ProjectsCreateResponse", "ProjectsImportTasksResponse", "ProjectsListResponse", "exports"]



# This file was auto-generated by Fern from our API Definition.

import typing

ProjectSampling = typing.Union[
    typing.Literal["Sequential sampling", "Uniform sampling", "Uncertainty sampling"], typing.Any
]
